import os
from taskinit import casalog, ms, qa
import numpy as np
from suncasa.utils import helioimage2fits as hf
import shutil
from functools import partial
from time import time
import glob


def clean_iter(tim, vis, imageprefix, imagesuffix,
               twidth, doreg, usephacenter, reftime, ephem, msinfo, toTb, overwrite,
               selectdata, field, spw,
               uvrange, antenna, scan, observation, intent, datacolumn, imsize, cell, phasecenter, stokes,
               projection, startmodel, specmode, reffreq, nchan, start, width, outframe, veltype, restfreq,
               interpolation, gridder, facets, chanchunks, wprojplanes, vptable, usepointing, mosweight, aterm, psterm,
               wbawp, conjbeams, cfcache, computepastep, rotatepastep, pblimit, normtype, deconvolver, scales, nterms,
               smallscalebias, restoration, restoringbeam, pbcor, outlierfile, weighting, robust, npixels, uvtaper,
               niter, gain, threshold, nsigma, cycleniter, cyclefactor, minpsffraction, maxpsffraction, interactive,
               usemask, mask, pbmask, sidelobethreshold, noisethreshold, lownoisethreshold, negativethreshold,
               smoothfactor, minbeamfrac, cutthreshold, growiterations, dogrowprune, minpercentchange, verbose, restart,
               savemodel, calcres, calcpsf, parallel, tmpdir, btidx):
    from tclean_cli import tclean_cli as tclean
    from split_cli import split_cli as split
    bt = btidx  # 0
    if bt + twidth < len(tim) - 1:
        et = btidx + twidth - 1
    else:
        et = len(tim) - 1

    if bt == 0:
        bt_d = tim[bt] - ((tim[bt + 1] - tim[bt]) / 2)
    else:
        bt_d = tim[bt] - ((tim[bt] - tim[bt - 1]) / 2)
    if et == (len(tim) - 1) or et == -1:
        et_d = tim[et] + ((tim[et] - tim[et - 1]) / 2)
    else:
        et_d = tim[et] + ((tim[et + 1] - tim[et]) / 2)

    timerange = qa.time(qa.quantity(bt_d, 's'), prec=9, form='ymd')[0] + '~' + \
                qa.time(qa.quantity(et_d, 's'), prec=9, form='ymd')[0]
    btstr = qa.time(qa.quantity(bt_d, 's'), prec=9, form='fits')[0]
    etstr = qa.time(qa.quantity(et_d, 's'), prec=9, form='fits')[0]
    print 'cleaning timerange: ' + timerange

    image0 = btstr.replace(':', '').replace('-', '')
    imname = imageprefix + image0 + imagesuffix
    ms_tmp = tmpdir + image0 + '.ms'
    print('checkpoint 1')
    # split(vis=vis, outputvis=ms_tmp, field=field, scan=scan, antenna=antenna, timerange=timerange,
    #       datacolumn=datacolumn)
    ms.open(vis)
    print('checkpoint 1-1')
    ms.split(ms_tmp,field=field, scan=scan, baseline=antenna, time=timerange,whichcol=datacolumn)
    print('checkpoint 1-2')
    ms.close()
    print('checkpoint 2')
    if overwrite or (len(glob.glob(imname + '*')) == 0):
        os.system('rm -rf {}*'.format(imname))
        try:
            tclean(vis=ms_tmp, selectdata=selectdata, field=field, spw=spw, timerange=timerange, uvrange=uvrange,
                   antenna=antenna, scan=scan, observation=observation, intent=intent, datacolumn=datacolumn,
                   imagename=imname, imsize=imsize, cell=cell, phasecenter=phasecenter, stokes=stokes,
                   projection=projection, startmodel=startmodel, specmode=specmode, reffreq=reffreq, nchan=nchan,
                   start=start, width=width, outframe=outframe, veltype=veltype, restfreq=restfreq,
                   interpolation=interpolation, gridder=gridder, facets=facets, chanchunks=chanchunks,
                   wprojplanes=wprojplanes, vptable=vptable, usepointing=usepointing, mosweight=mosweight, aterm=aterm,
                   psterm=psterm, wbawp=wbawp, conjbeams=conjbeams, cfcache=cfcache, computepastep=computepastep,
                   rotatepastep=rotatepastep, pblimit=pblimit, normtype=normtype, deconvolver=deconvolver,
                   scales=scales, nterms=nterms, smallscalebias=smallscalebias, restoration=restoration,
                   restoringbeam=restoringbeam, pbcor=pbcor, outlierfile=outlierfile, weighting=weighting,
                   robust=robust, npixels=npixels, uvtaper=uvtaper, niter=niter, gain=gain, threshold=threshold,
                   nsigma=nsigma, cycleniter=cycleniter, cyclefactor=cyclefactor, minpsffraction=minpsffraction,
                   maxpsffraction=maxpsffraction, interactive=interactive, usemask=usemask, mask=mask, pbmask=pbmask,
                   sidelobethreshold=sidelobethreshold, noisethreshold=noisethreshold,
                   lownoisethreshold=lownoisethreshold, negativethreshold=negativethreshold, smoothfactor=smoothfactor,
                   minbeamfrac=minbeamfrac, cutthreshold=cutthreshold, growiterations=growiterations,
                   dogrowprune=dogrowprune, minpercentchange=minpercentchange, verbose=verbose, restart=restart,
                   savemodel=savemodel, calcres=calcres, calcpsf=calcpsf, parallel=parallel)
            print('checkpoint 3')
            clnjunks = ['.flux', '.mask', '.model', '.psf', '.residual', '.pb', '.sumwt', '.image.pbcor']
            for clnjunk in clnjunks:
                if os.path.exists(imname + clnjunk):
                    shutil.rmtree(imname + clnjunk)
        except:
            print('error in cleaning image: ' + btstr)
            return [False, btstr, etstr, '']
    else:
        print imname + ' exists. Clean task aborted.'

    if doreg and not os.path.isfile(imname + '.fits'):
        # ephem.keys()
        # msinfo.keys()
        try:
            # check if ephemfile and msinfofile exist
            if not ephem:
                print("ephemeris info does not exist, querying from JPL Horizons on the fly")
                ephem = hf.read_horizons(vis=vis)
            if not msinfo:
                print("ms info not provided, generating one on the fly")
                msinfo = hf.read_msinfo(vis)
            hf.imreg(vis=vis, ephem=ephem, msinfo=msinfo, timerange=timerange, reftime=reftime,
                     imagefile=imname + '.image', fitsfile=imname + '.fits',
                     toTb=toTb, scl100=False, usephacenter=usephacenter)
            if os.path.exists(imname + '.fits'):
                shutil.rmtree(imname + '.image')
                return [True, btstr, etstr, imname + '.fits']
            else:
                return [False, btstr, etstr, '']
        except:
            print('error in registering image: ' + btstr)
            return [False, btstr, etstr, imname + '.image']
    else:
        if os.path.exists(imname + '.image'):
            return [True, btstr, etstr, imname + '.image']
        else:
            return [False, btstr, etstr, '']


def ptclean2(vis, imageprefix, imagesuffix, ncpu, twidth, doreg, usephacenter, reftime, toTb, overwrite, selectdata,
             field, spw, timerange, uvrange, antenna, scan, observation, intent, datacolumn, imsize, cell, phasecenter,
             stokes, projection, startmodel, specmode, reffreq, nchan, start, width, outframe, veltype, restfreq,
             interpolation, gridder, facets, chanchunks, wprojplanes, vptable, usepointing, mosweight, aterm, psterm,
             wbawp, conjbeams, cfcache, computepastep, rotatepastep, pblimit, normtype, deconvolver, scales, nterms,
             smallscalebias, restoration, restoringbeam, pbcor, outlierfile, weighting, robust, npixels, uvtaper, niter,
             gain, threshold, nsigma, cycleniter, cyclefactor, minpsffraction, maxpsffraction, interactive, usemask,
             mask, pbmask, sidelobethreshold, noisethreshold, lownoisethreshold, negativethreshold, smoothfactor,
             minbeamfrac, cutthreshold, growiterations, dogrowprune, minpercentchange, verbose, restart, savemodel,
             calcres, calcpsf, parallel):
    if not (type(ncpu) is int):
        casalog.post('ncpu should be an integer')
        ncpu = 8

    if doreg:
        # check if ephem and msinfo exist. If not, generate one on the fly
        try:
            ephem = hf.read_horizons(vis=vis)
        except ValueError:
            print("error in obtaining ephemeris")
        try:
            msinfo = hf.read_msinfo(vis)
        except ValueError:
            print("error in getting ms info")
    else:
        ephem = None
        msinfo = None

    if imageprefix:
        workdir = os.path.dirname(imageprefix)
    else:
        workdir = './'
    tmpdir = workdir + '/tmp/'
    if not os.path.exists(tmpdir):
        os.makedirs(tmpdir)
    # get number of time pixels
    ms.open(vis)
    ms.selectinit()
    timfreq = ms.getdata(['time', 'axis_info'], ifraxis=True)
    tim = timfreq['time']
    ms.close()

    if twidth < 1:
        casalog.post('twidth less than 1. Change to 1')
        twidth = 1

    if twidth > len(tim):
        casalog.post(
            'twidth greater than # of time pixels in the dataset. Change to the timerange of the entire dateset')
        twidth = len(tim)
    # find out the start and end time index according to the parameter timerange
    # if not defined (empty string), use start and end from the entire time of the ms
    if not timerange:
        btidx = 0
        etidx = len(tim) - 1
    else:
        try:
            (tstart, tend) = timerange.split('~')
            bt_s = qa.convert(qa.quantity(tstart, 's'), 's')['value']
            et_s = qa.convert(qa.quantity(tend, 's'), 's')['value']
            # only time is given but not date, add the date (at 0 UT) from the first record
            if bt_s < 86400. or et_s < 86400.:
                bt_s += np.fix(qa.convert(qa.quantity(tim[0], 's'), 'd')['value']) * 86400.
                et_s += np.fix(qa.convert(qa.quantity(tim[0], 's'), 'd')['value']) * 86400.
            btidx = np.argmin(np.abs(tim - bt_s))
            etidx = np.argmin(np.abs(tim - et_s))
            # make the indice back to those bracket by the timerange
            if tim[btidx] < bt_s:
                btidx += 1
            if tim[etidx] > et_s:
                etidx -= 1
            if etidx <= btidx:
                print "ending time must be greater than starting time"
                print "reinitiating to the entire time range"
                btidx = 0
                etidx = len(tim) - 1
        except ValueError:
            print "keyword 'timerange' has a wrong format"

    btstr = qa.time(qa.quantity(tim[btidx], 's'), prec=9, form='fits')[0]
    etstr = qa.time(qa.quantity(tim[etidx], 's'), prec=9, form='fits')[0]

    iterable = range(btidx, etidx + 1, twidth)
    print 'First time pixel: ' + btstr
    print 'Last time pixel: ' + etstr
    print str(len(iterable)) + ' images to clean...'

    res = []
    # partition
    clnpart = partial(clean_iter, tim, vis, imageprefix, imagesuffix,
                      twidth, doreg, usephacenter, reftime, ephem, msinfo, toTb, overwrite, selectdata, field, spw,
                      uvrange, antenna, scan, observation, intent, datacolumn, imsize, cell, phasecenter, stokes,
                      projection, startmodel, specmode, reffreq, nchan, start, width, outframe, veltype, restfreq,
                      interpolation, gridder, facets, chanchunks, wprojplanes, vptable, usepointing, mosweight, aterm,
                      psterm, wbawp, conjbeams, cfcache, computepastep, rotatepastep, pblimit, normtype, deconvolver,
                      scales, nterms, smallscalebias, restoration, restoringbeam, pbcor, outlierfile, weighting, robust,
                      npixels, uvtaper, niter, gain, threshold, nsigma, cycleniter, cyclefactor, minpsffraction,
                      maxpsffraction, interactive, usemask, mask, pbmask, sidelobethreshold, noisethreshold,
                      lownoisethreshold, negativethreshold, smoothfactor, minbeamfrac, cutthreshold, growiterations,
                      dogrowprune, minpercentchange, verbose, restart, savemodel, calcres, calcpsf, parallel, tmpdir)
    timelapse = 0
    t0 = time()
    # parallelization
    if ncpu > 1:
        import multiprocessing as mprocs
        casalog.post('Perform clean in parallel ...')
        print('Perform clean in parallel ...')
        pool = mprocs.Pool(ncpu)
        res = pool.map(clnpart, iterable)
        pool.close()
        pool.join()
    else:
        casalog.post('Perform clean in single process ...')
        print('Perform clean in single process ...')
        for i in iterable:
            res.append(clnpart(i))

    t1 = time()
    timelapse = t1 - t0
    print 'It took %f secs to complete' % timelapse
    # repackage this into a single dictionary
    results = {'Succeeded': [], 'BeginTime': [], 'EndTime': [], 'ImageName': []}
    for r in res:
        results['Succeeded'].append(r[0])
        results['BeginTime'].append(r[1])
        results['EndTime'].append(r[2])
        results['ImageName'].append(r[3])

    if os.path.exists(tmpdir):
        os.system('rm -rf ' + tmpdir)

    return results



# import os
# import matplotlib as mpl
import matplotlib.pyplot as plt
# import matplotlib.colorbar as colorbar
# import sunpy.cm.cm as cm  ## to bootstrap sdoaia color map
# import matplotlib.cm as cm
# import matplotlib.colors as colors
import astropy.units as u
# from astropy.io import fits
# import matplotlib.dates as mdates
# from astropy.time import Time
from sunpy import map as smap
# import matplotlib.gridspec as gridspec
# import numpy.ma as ma
# import matplotlib.patches as patches
# from suncasa.utils import stackplot as stp
# from IPython import embed
# from astropy.coordinates import SkyCoord
import numpy as np
from suncasa.utils import DButil
import warnings
import matplotlib.patches as patches
import numpy.ma as ma


class Sunmap():

    def __init__(self, sunmap, aia=False):
        if aia:
            try:
                sunmap = DButil.normalize_aiamap(sunmap)
            except:
                pass
            data = sunmap.data
            data[data < 1.0] = 1.0
            self.sunmap = smap.Map(data, sunmap.meta)
        else:
            self.sunmap = sunmap

    def map2wcsgrids(self, sunpymap=None, cell=False):
        '''

        :param sunpymap:
        :param cell: if True, return the coordinates of the pixel centers. if False, return the coordinates of the pixel boundaries
        :return:
        '''
        # embed()
        import astropy.units as u
        from suncasa.utils import stputils as stpu
        if sunpymap is None:
            sunpymap = self.sunmap
        ny, nx = sunpymap.data.shape
        x0, x1, y0, y1 = stpu.get_map_corner_coord(sunpymap).value

        dx = sunpymap.scale.axis1.to(u.arcsec / u.pix).value
        dy = sunpymap.scale.axis2.to(u.arcsec / u.pix).value

        if cell:
            mapx, mapy = np.linspace(x0, x1, nx) - dx / 2.0, np.linspace(y0, y1, ny) - dy / 2.0
            mapx = np.tile(mapx, ny).reshape(ny, nx)
            mapy = np.tile(mapy, nx).reshape(nx, ny).transpose()
        else:
            nx += 1
            ny += 1
            mapx, mapy = np.linspace(x0 - dx, x1 + dx, nx), np.linspace(y0 - dy, y1 + dy, ny)
            mapx = np.tile(mapx, ny).reshape(ny, nx)
            mapy = np.tile(mapy, nx).reshape(nx, ny).transpose()
        return mapx, mapy

    def get_map_extent(self, sunpymap=None, rot=0, rangereverse=False):
        from suncasa.utils import stputils as stpu
        if sunpymap is None:
            sunpymap = self.sunmap
        rot = rot % 360
        x0, x1, y0, y1 = stpu.get_map_corner_coord(sunpymap).value
        if rot == 90:
            extent = np.array([y1, y0, x0, x1])
            extent = extent - np.array([sunpymap.scale.axis2.value] * 2 + [sunpymap.scale.axis1.value] * 2) / 2.0
        elif rot == 180:
            extent = np.array([x1, x0, y1, y0])
            extent = extent - np.array([sunpymap.scale.axis1.value] * 2 + [sunpymap.scale.axis2.value] * 2) / 2.0
        elif rot == 270:
            extent = np.array([y0, y1, x1, x0])
            extent = extent - np.array([sunpymap.scale.axis1.value] * 2 + [sunpymap.scale.axis2.value] * 2) / 2.0
        else:
            extent = np.array([x0, x1, y0, y1])
            extent = extent - np.array([sunpymap.scale.axis1.value] * 2 + [sunpymap.scale.axis2.value] * 2) / 2.0
        if rangereverse:
            x0, x1, y0, y1 = extent
            extent = -np.array([x1, x0, y1, y0])
        return extent

    def imshow(self, axes=None, rot=0, rangereverse=False, maskon=False, image_enhance=False, **kwargs):
        '''
        :param sunpymap:
        :param axes:
        :param rot: rotation angle in degrees counter-clockwise. Must be an integer multiple of 90.
        :param kwargs:
        :return:
        '''
        sunpymap = self.sunmap
        if axes is None:
            axes = plt.subplot()
        rot = rot % 360
        if rot == 0:
            imdata = sunpymap.data
        elif rot == 90:
            imdata = sunpymap.data.transpose()[:, ::-1]
        elif rot == 180:
            imdata = sunpymap.data[::-1, ::-1]
        elif rot == 270:
            imdata = sunpymap.data.transpose()[::-1, :]
        else:
            warnings.warn('rot must be an integer multiple of 90. rot not implemented!')
            imdata = sunpymap.data
            rot = 0
        extent = self.get_map_extent(rot=rot, rangereverse=rangereverse)

        if maskon:
            if isinstance(maskon, bool):
                imdataplt = ma.masked_invalid(imdata)
                immask = imdataplt.mask
            elif isinstance(maskon, dict):
                if 'masked_equal' in maskon.keys():
                    imdataplt = ma.masked_equal(imdata, maskon['masked_equal'])
                elif 'masked_greater' in maskon.keys():
                    imdataplt = ma.masked_greater(imdata, maskon['masked_greater'])
                elif 'masked_less' in maskon.keys():
                    imdataplt = ma.masked_less(imdata, maskon['masked_less'])
                elif 'masked_greater_equal' in maskon.keys():
                    imdataplt = ma.masked_greater_equal(imdata, maskon['masked_greater_equal'])
                elif 'masked_less_equal' in maskon.keys():
                    imdataplt = ma.masked_less_equal(imdata, maskon['masked_less_equal'])
                elif 'masked_outside' in maskon.keys():
                    v1, v2 = maskon['masked_outside']
                    imdataplt = ma.masked_outside(imdata, v1, v2)
                elif 'masked_inside' in maskon.keys():
                    v1, v2 = maskon['masked_inside']
                    imdataplt = ma.masked_inside(imdata, v1, v2)
                elif 'masked_invalid' in maskon.keys():
                    imdataplt = ma.masked_invalid(imdata)
                else:
                    raise ValueError('maskon key wrong.')
                immask = imdataplt.mask
            else:
                raise TypeError('maskon must be bool or dict type.')
        else:
            imdataplt = imdata.copy()

        if image_enhance:
            dmax = np.nanmax(imdataplt)
            dmin = np.nanmin(imdataplt)
            from skimage.exposure import equalize_adapthist
            if isinstance(image_enhance, dict):
                imdataplt = equalize_adapthist(imdataplt, **image_enhance) * (dmax - dmin) + dmin
            else:
                imdataplt = equalize_adapthist(imdataplt) * (dmax - dmin) + dmin

        if maskon:
            imdataplt = ma.masked_array(imdataplt, immask)

        if isinstance(axes, list):
            ims = []
            for ax in axes:
                im = ax.imshow(imdataplt, extent=extent, origin='lower', **kwargs)
                ims.append(im)
                if rot == 0:
                    ax.set_xlabel('Solar X [arcsec]')
                    ax.set_ylabel('Solar Y [arcsec]')
                elif rot == 90:
                    ax.set_xlabel('-Solar Y [arcsec]')
                    ax.set_ylabel('Solar X [arcsec]')
                elif rot == 180:
                    ax.set_xlabel('-Solar X [arcsec]')
                    ax.set_ylabel('-Solar Y [arcsec]')
                elif rot == 270:
                    ax.set_xlabel('Solar Y [arcsec]')
                    ax.set_ylabel('-Solar X [arcsec]')
            return ims
        else:
            im = axes.imshow(imdataplt, extent=extent, origin='lower', **kwargs)

            if rot == 0:
                axes.set_xlabel('Solar X [arcsec]')
                axes.set_ylabel('Solar Y [arcsec]')
            elif rot == 90:
                axes.set_xlabel('-Solar Y [arcsec]')
                axes.set_ylabel('Solar X [arcsec]')
            elif rot == 180:
                axes.set_xlabel('-Solar X [arcsec]')
                axes.set_ylabel('-Solar Y [arcsec]')
            elif rot == 270:
                axes.set_xlabel('Solar Y [arcsec]')
                axes.set_ylabel('-Solar X [arcsec]')
            return im

    def contour(self, axes=None, rot=0, mapx=None, mapy=None, rangereverse=False, **kwargs):
        sunpymap = self.sunmap
        if axes is None:
            axes = plt.subplot()
        rot = rot % 360
        if (mapx is None) or (mapy is None):
            if rot == 0:
                mapx, mapy = self.map2wcsgrids(cell=True)
            elif rot == 90:
                mapy, mapx = self.map2wcsgrids(cell=True)
            elif rot == 180:
                mapx, mapy = self.map2wcsgrids(cell=True)
            elif rot == 270:
                mapy, mapx = self.map2wcsgrids(cell=True)

        if isinstance(axes, list):
            ims = []
            for ax in axes:
                im = ax.contour(mapx, mapy, sunpymap.data, **kwargs)
                ims.append(im)
                extent = self.get_map_extent(rot=rot, rangereverse=rangereverse)
                ax.set_xlim(extent[:2])
                ax.set_ylim(extent[2:])
            return ims
        else:
            im = axes.contour(mapx, mapy, sunpymap.data, **kwargs)
            extent = self.get_map_extent(rot=rot, rangereverse=rangereverse)
            axes.set_xlim(extent[:2])
            axes.set_ylim(extent[2:])
            return im

    def contourf(self, axes=None, rot=0, mapx=None, mapy=None, rangereverse=False, **kwargs):
        sunpymap = self.sunmap
        if axes is None:
            axes = plt.subplot()
        rot = rot % 360
        if (mapx is None) or (mapy is None):
            if rot == 0:
                mapx, mapy = self.map2wcsgrids(cell=True)
            elif rot == 90:
                mapy, mapx = self.map2wcsgrids(cell=True)
            elif rot == 180:
                mapx, mapy = self.map2wcsgrids(cell=True)
            elif rot == 270:
                mapy, mapx = self.map2wcsgrids(cell=True)

        if isinstance(axes, list):
            ims = []
            for ax in axes:
                im = ax.contourf(mapx, mapy, sunpymap.data, **kwargs)
                ims.append(im)
                extent = self.get_map_extent(rot=rot, rangereverse=rangereverse)
                ax.set_xlim(extent[:2])
                ax.set_ylim(extent[2:])
            return ims
        else:
            im = axes.contourf(mapx, mapy, sunpymap.data, **kwargs)
            extent = self.get_map_extent(rot=rot, rangereverse=rangereverse)
            axes.set_xlim(extent[:2])
            axes.set_ylim(extent[2:])
            return im

    def draw_limb(self, axes=None, rangereverse=False, **kwargs):
        if 'c' not in kwargs and 'color' not in kwargs:
            kwargs['c'] = 'w'
        if 'ls' not in kwargs and 'linestyle' not in kwargs:
            kwargs['ls'] = 'solid'
        sunpymap = self.sunmap
        if axes is None:
            axes = plt.gca()

        rsun = sunpymap.rsun_obs
        phi = np.linspace(-180, 180, num=181) * u.deg
        x = np.cos(phi) * rsun
        y = np.sin(phi) * rsun
        if isinstance(axes, list):
            ims = []
            for ax in axes:
                ax.set_autoscale_on(False)
                im = ax.plot(x, y, **kwargs)
                ims.append(im)
            return ims
        else:
            axes.set_autoscale_on(False)
            im = axes.plot(x, y, **kwargs)
            return im

    def draw_grid(self, axes=None, rot=0, grid_spacing=None, **kwargs):
        sunpymap = self.sunmap
        if grid_spacing is None:
            grid_spacing = 15 * u.deg

        def hgs2hcc(rsun, lon, lat, B0, L0):
            lon_L0 = lon - L0
            x = rsun * np.cos(lat) * np.sin(lon)
            y = rsun * (np.sin(lat) * np.cos(B0) - np.cos(lat) * np.cos(lon_L0) * np.sin(B0))
            z = rsun * (np.sin(lat) * np.sin(B0) + np.cos(lat) * np.cos(lon_L0) * np.cos(B0))
            return x, y, z

        def hcc2hpc(x, y, z, dsun):
            d = np.sqrt(x ** 2 + y ** 2 + (dsun - z) ** 2)
            Tx = np.arctan2(x, dsun - z)
            Ty = np.arcsin(y / d)
            return Tx, Ty

        if 'c' not in kwargs and 'color' not in kwargs:
            kwargs['c'] = 'w'
        if 'ls' not in kwargs and 'linestyle' not in kwargs:
            kwargs['ls'] = 'dotted'
        dsun = sunpymap.dsun
        rsun = sunpymap.rsun_meters
        if axes is None:
            axes = plt.gca()
        im = []
        b0 = sunpymap.heliographic_latitude.to(u.deg)
        l0 = sunpymap.heliographic_longitude.to(u.deg)
        hg_longitude_deg = np.linspace(-90, 90, num=91) * u.deg
        hg_latitude_deg = np.arange(0, 90, grid_spacing.to(u.deg).value)
        hg_latitude_deg = np.hstack([-hg_latitude_deg[1:][::-1], hg_latitude_deg]) * u.deg
        for lat in hg_latitude_deg:
            c = hgs2hcc(rsun, hg_longitude_deg, lat * np.ones(91), b0, l0)
            coords = hcc2hpc(c[0], c[1], c[2], dsun)
            if rot in [90, 270]:
                coords_ = [coords[1], coords[0]]
            else:
                coords_ = coords
            if isinstance(axes, list):
                for ax in axes:
                    im += ax.plot(coords_[0].to(u.arcsec), coords_[1].to(u.arcsec), **kwargs)
            else:
                im += axes.plot(coords_[0].to(u.arcsec), coords_[1].to(u.arcsec), **kwargs)

        hg_longitude_deg = np.arange(0, 90, grid_spacing.to(u.deg).value)
        hg_longitude_deg = np.hstack([-hg_longitude_deg[1:][::-1], hg_longitude_deg]) * u.deg
        hg_latitude_deg = np.linspace(-90, 90, num=91) * u.deg

        for lon in hg_longitude_deg:
            c = hgs2hcc(rsun, lon * np.ones(91), hg_latitude_deg, b0, l0)
            coords = hcc2hpc(c[0], c[1], c[2], dsun)
            if rot in [90, 270]:
                coords_ = [coords[1], coords[0]]
            else:
                coords_ = coords
            if isinstance(axes, list):
                for ax in axes:
                    im += ax.plot(coords_[0].to(u.arcsec), coords_[1].to(u.arcsec), **kwargs)
            else:
                im += axes.plot(coords_[0].to(u.arcsec), coords_[1].to(u.arcsec), **kwargs)
        return im

    def draw_rectangle(self, bottom_left, width, height, axes=None, **kwargs):
        if 'ec' not in kwargs and 'edgecolor' not in kwargs:
            kwargs['ec'] = 'w'
        if 'ls' not in kwargs and 'linestyle' not in kwargs:
            kwargs['ls'] = 'solid'
        if 'fill' not in kwargs:
            kwargs['fill'] = False
        if axes is None:
            axes = plt.gca()

        if isinstance(axes, list):
            ims = []
            for ax in axes:
                ax.set_autoscale_on(False)
                im = ax.add_patch(patches.Rectangle(bottom_left, width, height, **kwargs))
                ims.append(im)
            return ims
        else:
            axes.set_autoscale_on(False)
            im = axes.add_patch(patches.Rectangle(bottom_left, width, height, **kwargs))
            return im

    def imshow_RGB(self, maps, axes=None, returndataonly=False, rangereverse=False):
        from scipy import ndimage
        from astropy.coordinates import SkyCoord
        mapR = maps[0]
        znewR = mapR.data
        aiamapx, aiamapy = self.map2wcsgrids(sunpymap=mapR, cell=False)
        mapG = maps[1]
        XX, YY = mapG.data_to_pixel(SkyCoord(aiamapx * u.arcsec, aiamapy * u.arcsec, frame=mapG.coordinate_frame))
        znewG = ndimage.map_coordinates(mapG.data, [YY, XX], order=1)
        mapB = maps[2]
        XX, YY = mapB.data_to_pixel(SkyCoord(aiamapx * u.arcsec, aiamapy * u.arcsec, frame=mapB.coordinate_frame))
        znewB = ndimage.map_coordinates(mapB.data, [YY, XX], order=1)

        znewR = np.sqrt(znewR)
        znewG = np.sqrt(znewG)
        znewB = np.sqrt(znewB)

        vmax, vmin = np.sqrt(5000), np.sqrt(10)
        # clrange=DButil.sdo_aia_scale_dict(304)
        znewR[znewR > vmax] = vmax
        znewR[znewR < vmin] = vmin
        # clrange=DButil.sdo_aia_scale_dict(94)
        vmax, vmin = np.sqrt(20000), np.sqrt(200)
        znewG[znewG > vmax] = vmax
        znewG[znewG < vmin] = vmin
        # clrange=DButil.sdo_aia_scale_dict(211)
        vmax, vmin = np.sqrt(5000), np.sqrt(100)
        znewB[znewB > vmax] = vmax
        znewB[znewB < vmin] = vmin
        znewR = (znewR - np.nanmin(znewR)) / (np.nanmax(znewR) - np.nanmin(znewR))
        znewG = (znewG - np.nanmin(znewG)) / (np.nanmax(znewG) - np.nanmin(znewG))
        znewB = (znewB - np.nanmin(znewB)) / (np.nanmax(znewB) - np.nanmin(znewB))
        # znew1 = np.sqrt(znew1)
        # znew2 = np.sqrt(znew2)
        # imshow(np.sqrt(np.stack([znew0, znew1, znew2], axis=-1)), extent=list(aiamap.xrange.value) + list(aiamap.yrange.value),origin='lower')
        if returndataonly:
            return np.stack([znewR, znewG, znewB], axis=-1)
        else:
            if axes:
                pass
            else:
                axes = plt.subplot()
            extent = self.get_map_extent(sunpymap=mapR, rangereverse=rangereverse)
            if isinstance(axes, list):
                ims = []
                for ax in axes:
                    im = ax.imshow(np.stack([znewR, znewG, znewB], axis=-1),
                                   extent=extent, origin='lower')
                    ims.append(im)
                return ims
            else:
                return axes.imshow(np.stack([znewR, znewG, znewB], axis=-1),
                                   extent=extent, origin='lower')



from astropy.io import fits
import astropy.table
from astropy.time import Time
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from matplotlib.dates import DateFormatter
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable


def get_dspec(filename, doplot=False, vmax=None, vmin=None, norm=None, cmap=None):
    """
    Read EOVSA Dynamic Spectrum FITS file <filename> and return a spectrogram dictionary.
    Optionally show an overview plot if doplot switch is set.

    Example:
    --------
    >>> from suncasa.eovsa import eovsa_dspec as ds
    >>> from astropy.time import Time
    >>> from matplotlib.colors import LogNorm
    ## Read EOVSA Dynamic Spectrum FITS file <filename>
    >>> filename = 'EOVSA_TPall_20170713.fts'
    >>> s = ds.get_dspec(filename, doplot=True, cmap='gist_heat', norm=LogNorm(vmax=2.1e3, vmin=40))
    ## To access the data in the spectrogram object, use
    >>> spec = s['spectrogram']                    ## (Array of amplitudes in SFU, of size nfreq,ntimes)
    >>> fghz = s['spectrum_axis']                  ## (Array of frequencies in GHz, of size nfreq)
    >>> tim = Time(s['time_axis'], format='mjd')   ## (Array of UT times in astropy.time object, of size ntimes)

    Parameters
    ----------
    filename :   filename of the spectrogram fits file

    doplot : Boolean, optional

    vmin, vmax : scalar, optional
    When using scalar data and no explicit norm, vmin and vmax
    define the data range that the colormap covers. By default,
    the colormap covers the complete value range of the supplied data.
    vmin, vmax are ignored if the norm parameter is used.

    norm : `matplotib.colors Normalization object` or str, optional
    The Normalize instance used to scale scalar data to the [0, 1]
    range before mapping to colors using cmap. By default, a linear
    scaling mapping the lowest value to 0 and the highest to 1 is used.
    This parameter is ignored for RGB(A) data.

    cmap : `matplotib.colors.Colormap` or str
    A colormap instance or the name of a registered colormap.

    Returns
    -------
    spectrogram : dictionary
    """

    hdulist = fits.open(filename)
    spec = hdulist[0].data
    fghz = np.array(astropy.table.Table(hdulist[1].data)['sfreq'])
    tim = astropy.table.Table(hdulist[2].data)
    tmjd= np.array(tim['mjd']) + np.array(tim['time']) / 24. / 3600 / 1000
    tim = Time(tmjd, format='mjd')
    timplt = tim.plot_date
    ntim = len(timplt)
    nfreq = len(fghz)
    if doplot:
        fig, ax = plt.subplots(figsize=(7, 4))
        # if vmax is None:
        #     vmax = np.nanmax(spec)
        # if vmin is None:
        #     vmin = 0.1
        # if norm is None:
        #     norm = colors.Normalize(vmax=vmax, vmin=vmin)
        pcm = ax.pcolormesh(timplt, fghz, spec, norm=norm, vmax=vmax, vmin=vmin, cmap=cmap,rasterized='True')
        ax.xaxis_date()
        ax.xaxis.set_major_formatter(DateFormatter("%H:%M"))
        ax.set_ylim(fghz[0], fghz[-1])
        ax.set_xlim(tim[0].plot_date, tim[-1].plot_date)
        ax.set_xlabel('Time [UT]')
        ax.set_ylabel('Frequency [GHz]')
        ax.set_title('EOVSA Dynamic Spectrum for ' + tim[0].datetime.strftime('%Y-%m-%d'))
        divider = make_axes_locatable(ax)
        cax_spec = divider.append_axes('right', size='1.5%', pad=0.05)
        clb_spec = plt.colorbar(pcm, ax=ax, cax=cax_spec, label='Flux density [sfu]')
        cmap = pcm.get_cmap()
        bkg_facecolor=cmap(0.0)
        ax.set_facecolor(bkg_facecolor)
        dtim = np.diff(tmjd)
        idxs_tgap = np.where(dtim*24*60 >= 0.5)[0]
        if len(idxs_tgap)>0:
            for idx in idxs_tgap:
                if idx+1 < len(tmjd):
                    ax.axvspan(tim[idx].plot_date,tim[idx+1].plot_date,facecolor=bkg_facecolor,edgecolor='none')



        def format_coord(x, y):
            col = np.argmin(np.absolute(timplt - x))
            row = np.argmin(np.absolute(fghz - y))
            if col >= 0 and col < ntim and row >= 0 and row < nfreq:
                timstr = tim[col].isot
                flux = spec[row, col]
                return 'time {0}, freq = {1:.3f} GHz, flux = {2:.2f} sfu'.format(timstr, y, flux)
            else:
                return 'x = {0}, y = {1:.3f}'.format(x, y)

        ax.format_coord = format_coord
        fig.tight_layout()
        plt.show()
    return {'spectrogram': spec, 'spectrum_axis': fghz, 'time_axis': tmjd}



## build suncasa tasks on inti
## export PATH="$PATH:/inti/software/casa-6.2.0-124/bin/"
## [ref]: https://casadocs.readthedocs.io/en/latest/api/casashell/buildmytasks.html#
## using buildmytasks requires a full installation of CASA. https://casadocs.readthedocs.io/en/stable/notebooks/usingcasa.html

import os
from glob import glob

xmlfiles = glob('ptclean6*.xml')

for xmlfile in xmlfiles:
    # os.system("buildmytasks --upgrade {}".format(xmlfile)) ## this only need to be run once.
    os.system("buildmytasks --module suncasatasks {}".format(xmlfile))
